Fundamentals of Computer Music Final Project
Antonis Stampoulis (antonios.stampoulis@yale.edu)

In the second part of my project, I investigated the possibility of
enhancing the HasSound frontend with an arrow-based interface. The
main idea is that we can view the process of constructing signals as
the wiring together of SigExp's different constructors. For an
instrument, the resulting arrow will be of a type similar to SigExp
[CSnote] -> SigExp CSaudio. That's because we view an instrument as a
signal processing function that takes a list of inputs at note rate
(the p-fields), and produces an audio-rate signal.

In doing this, a fundamental problem arises. The normal arrow type
class defines a method to lift a normal function to the arrow type,
namely the method arr (or pure). This function has to have a type of
(a -> b) -> (arrow a b). In our case, where arrow a b will be
something similar to (SigExp a -> SigExp b), this function does not
make any sense, and cannot even be implemented! More precisely,
lifting a function with a type, say, Int -> Int to the arrow type
would have no meaning (since there are no SigExp Int's). While this
could not be a problem if we could handle lifting functions that do
make sense correctly, that also is impossible. We would like to create
arrows like (arrow CSnote CSnote), but that is not possible because
there is no function with type CSnote -> CSnote to lift. (That's
because CSnote is an empty type, and has no constructors). The
essence of this problem is that while we want to restrict ourselves
to the domain of SigExp's, the normal arrow type class requires
that our arrow type can handle any Haskell type.

Still, there is a way to solve this problem: instead of using the
normal arrow type class, we can define our own arrow type class
that doesn't have the same problem. This type class should be as
close to the original type class as possible. Note that while
this solves our problem, it means that we will not be able to
use the normal arrow syntax to write our instruments -- we would
have to create our own arrow preprocessor to do that.

The type class we define is a very slight departure from the original
one. It provides exactly the same methods, but takes two arguments: one
is the actual arrow type class, and the other one is the domain we restrict
ourselves to. This domain is given in the form of a type constructor, which
in our case will just be SigExp, and the methods' signatures are changed in
order to have them handle only types produced by that type constructor. The
definition of our new type class follows.

> class Arrow arrow domain where
>     arr    :: (domain x -> domain y) -> arrow (domain x) (domain y)
>     (>>>)  :: arrow (domain x) (domain y) -> arrow (domain y) (domain z) -> arrow (domain x) (domain z)
>     first  :: arrow (domain x) (domain y) -> arrow (domain (x, z)) (domain (y, z))
>     second :: arrow (domain x) (domain y) -> arrow (domain (z, x)) (domain (z, y))
>     (***)  :: arrow (domain x) (domain y) -> arrow (domain x') (domain y') -> arrow (domain (x, x')) (domain (y, y'))
>     (&&&)  :: arrow (domain x) (domain y) -> arrow (domain x)  (domain y') -> arrow (domain x) (domain (y, y'))

Now we will define the instance of this type class using the function
type constructor as the arrow type, and SigExp as our domain. Note that
we are doing the implementation first, to see what our SigExp's should
support, and then define SigExp's accordingly. We see that we will need
a pair constructor to form a signal that is a pair of two other signals,
and of course the appropriate deconstructors (First and Second).

> instance Arrow (->) SigExp where
>     arr = id
>     (>>>) f g = g . f
>     first  f xz = Pair (f (First xz)) (Second xz)
>     second f zx = Pair (First zx)     (f (Second zx))
>     (***)  f g xx' = Pair (f $ First xx') (g $ Second xx')
>     (&&&)  f g x   = Pair (f x) (g x) 

What about recursive signals? To support these, we also need to
implement the ArrowLoop type class. As for Arrow, we will define our
own ArrowLoop type class. But, we will need more changes to it: in
order to compile recursive signals, we need to be able to tell what's
the type of the recursive variable. That information will be passed as
a value-level entity to the constructor that handles recursive
signals. The type of that entity must reflect the type of the
recursive variable (so that we can do strict typechecking), and thus
must be an instance of a GADT, that describes all the allowed
recursive variable types. This GADT is called AllTypes, but in order
for our ArrowLoop class to be general enough, we make it a third
parameter to the class. We provide below the definition of the
AllTypes GADT, the definition of our ArrowLoop class and its instance
for (->), SigExp and AllTypes.

> data AllTypes a where
>     Tar :: AllTypes CSaudio
>     Tcr :: AllTypes CScontrol
>     Tnr :: AllTypes CSnote
>     Tunit :: AllTypes ()
>     Tpair :: AllTypes a -> AllTypes b -> AllTypes (a, b)

> class Arrow a b => ArrowLoop a b t where
>     loop :: t z -> a (b (x, z)) (b (y, z)) -> a (b x) (b y)

> instance ArrowLoop (->) SigExp AllTypes where
>     loop var f x =
>          Recurse var (\z -> let yz = f (Pair x z) in (First yz, Second yz))

In the above implementation of ArrowLoop, we can see that Recurse is
used in a different way than the previous part of this project (in
gadts.lhs). In this modified form, the recurse constructor takes a
description of the recursive variable's type and a function that
given the previous value of the recursive variable, calculates separately
the resulting signal, and the new value of the recursive variable.
The difference with the previous form is that the result of the Recurse
operator doesn't have to be the recursive variable itself. A graphical
representation of the two Recurse operators is this:

(simple recurse)
       _______ 
x ----|_______|---.----- x
    ^_____________|

(the recurse operator used here)

       _______ 
x ----|       |------.-- x
    ^ |_______|--- y |
    |________________|

We should note another thing: the Recurse operator should support any
type of recursive signal (both as input and output), and this now
includes paired signals! This means that a single Recurse operation
can have multiple inputs and multiple outputs.

Considering all these, we can now define our desired SigExp type, and
worry later about how we will actually translate it into CSound. As
before, we have to define the CSound rate-types and their subtyping
rules, the Function type, and the syntactic sugar for rate
constraints. But we will provide these later, since they are unchanged
from the previous part. The differences lie in SigExp, which we define
below.

> data SigExp a where

A subtle difference: the constant signals are actually defined as signal
processors -- the take a signal of any type, discard their input, and
provide a constant signal. This is done so that programming using the
arrow interface is more natural: we can pair a signal with a constant
using just the (&&&) operator.

>     ConstFloat   :: Float -> SigExp a -> SigExp CSconst
>     ConstInt     :: Int   -> SigExp a -> SigExp CSconst
>     TableNumber  :: Int   -> SigExp a -> SigExp CSconst

PFields are defined as before. In the arrow interface, they are not to
be used directly! PFields should be read from the input signal.

>     PField       :: Int   -> SigExp CSnote

Infix operators are defined as before, with the difference that they
take only one input: a SigExp that is a pair of two signals!

>     Infix        :: CSisSubtype a b => Function -> SigExp (a, b) -> SigExp b

Prefix and SigGen constructors are exactly the same as before. Functions
wrapping the SigGen constructor should convert to and from signal pairs
as needed.

>     Prefix       :: CSType a => Function -> SigExp a -> SigExp a
>     SigGen       :: Function -> EvalRate a -> [ SigExpAny ] -> SigExp a

The Recurse operator features the difference described above. It needs
a function that gets a variable, and produces two results: the actual resulting
signal and the updated value for the recursive variable. Also, the recursive
variable can be of any type, including a pair type!

>     Recurse      :: AllTypes var -> (SigExp var -> (SigExp res, SigExp var)) -> SigExp res

Now, we have to define the other types that we support in the SigExp
universe.  We provide a unit type (so that we have a value we can just
pass to a constant-signal processor to get the actual constant value),
a pair type and its deconstructors, and also a list type with the nth
function as deconstructor (instead of head/tail). Lists are given just
as a convenience type for defining the input; they are not supported
as deeply as pairs, so we cannot have a recursive variable of list
type.

>     Unit         :: SigExp ()
>     Pair         :: SigExp a -> SigExp b -> SigExp (a, b)
>     First        :: SigExp (a, b) -> SigExp a
>     Second       :: SigExp (a, b) -> SigExp b
>     Nil          :: SigExp [a]
>     Cons         :: SigExp a -> SigExp [a] -> SigExp [a]
>     Nth          :: Int -> SigExp [a] -> SigExp a

This is just used internally in the compilation process.

>     Var          :: Tree String -> SigExp a

And that's all there is to the SigExp data type! Let's now define the
types SigExp depends upon.

--------------------------------------------------------------
The following definitions are needed by the SigExp type, but
they are unchanged from the previous part, so you can skip
re-reading them.

> data CSconst
> data CSnote
> data CScontrol
> data CSaudio

> class CSisSubtype a b
> instance CSisSubtype CSconst CSconst
> instance CSisSubtype CSconst CSnote
> instance CSisSubtype CSconst CScontrol
> instance CSisSubtype CSconst CSaudio
> instance CSisSubtype CSnote  CSnote
> instance CSisSubtype CSnote  CScontrol
> instance CSisSubtype CSnote  CSaudio
> instance CSisSubtype CScontrol CScontrol
> instance CSisSubtype CScontrol CSaudio
> instance CSisSubtype CSaudio   CSaudio

> type Function = String

> data EvalRate a where
>     AR :: EvalRate CSaudio
>     CR :: EvalRate CScontrol
>     NR :: EvalRate CSnote

> data SigExpAny where
>     Hide :: forall a . SigExp a -> SigExpAny

> class CSConst   a
> instance CSConst CSconst

> class CSNote    a
> instance CSNote CSconst
> instance CSNote CSnote

> class CSControl a
> instance CSControl CSconst
> instance CSControl CSnote
> instance CSControl CScontrol

> class CSAudio   a
> instance CSAudio CSconst
> instance CSAudio CSnote
> instance CSAudio CScontrol
> instance CSAudio CSaudio

> class CSType    a
> instance CSType CSconst
> instance CSType CSnote
> instance CSType CScontrol
> instance CSType CSaudio

---------------------------------------------------------------------------



So now we can try our new SigExp out. We will define a recursive
signal that is actually defined using a pair as a recursive
variable. This depends on definitions of oscPure, line1 etc., which
are the same as in the previous part. Note that we are not using the
arrow operators yet.


> pureToneTable = TableNumber 1 Unit

> osc :: (CSAudio amp, CSAudio freq, CSConst table) =>
>        EvalRate a -> SigExp table -> SigExp amp -> SigExp freq -> SigExp a
> osc rate table amp freq = SigGen "oscil" rate [ Hide amp, Hide freq, Hide table ]

> oscPure :: (CSAudio amp, CSAudio freq) =>
>            SigExp amp -> SigExp freq -> SigExp CSaudio
> oscPure = osc AR pureToneTable

> line1 :: (CSNote ia, CSNote idur1, CSNote ib) =>
>          EvalRate a -> SigExp ia -> SigExp idur1 -> SigExp ib -> SigExp a
> line1 rate start dur end = SigGen "line" rate [ Hide start, Hide dur, Hide end ]

> delay1 :: (CSAudio sig, CSNote del) =>
>           SigExp del -> SigExp sig -> SigExp CSaudio
> delay1 del sig = SigGen "delay" AR [ Hide sig, Hide del ]

> signal :: SigExp CSaudio
> signal =
>     let freq = Prefix "cpspch" (PField 4)
>         amp  = Prefix "ampdb"  (PField 5)
>         env  = line1 CR (ConstFloat 0.0 Unit) (PField 3) (ConstFloat 1.0 Unit)
>         amp' = Infix "*" (Pair amp env)
>         freq' = Infix "+" (Pair freq (oscPure (ConstFloat 5.0 Unit) (Infix "*" (Pair (ConstFloat 5.0 Unit) (PField 3)))))
>         sig  = oscPure amp' freq'
>         sig' = Recurse (Tpair Tar Tar)
>                         (\ab -> let
>                                    a = First ab
>                                    b = Second ab
>                                    new = Pair (Infix "+" (Pair sig (delay1 (ConstFloat 0.311 Unit) a)))
>                                               (Infix "+" (Pair sig (delay1 (ConstFloat 0.512 Unit) b)))
>                                in
>                                (Infix "+" ab, new))
>     in  sig'


OK, so now let's try the arrow interface out. We provide wrapper
functions that can be used in the arrow setting.

> arrowUncurry :: (SigExp a -> SigExp b -> SigExp c) -> (SigExp (a,b) -> SigExp c)
> arrowUncurry f a = f (First a) (Second a)

> arrowUncurry3 :: (SigExp a -> SigExp b -> SigExp c -> SigExp d) -> (SigExp ((a,b),c) -> SigExp d)
> arrowUncurry3 f a = f (First (First a)) (Second (First a)) (Second a)

> oscPureArrow :: (CSAudio amp, CSAudio freq) => (SigExp (amp, freq) -> SigExp CSaudio)
> oscPureArrow = arrowUncurry oscPure

> lineArrow :: (CSNote ia, CSNote idur1, CSNote ib) => EvalRate a ->
>              (SigExp ((ia, idur1), ib) -> SigExp a)
> lineArrow a = arrowUncurry3 (line1 a)

> delayArrow :: (CSAudio sig, CSNote del) =>
>               (SigExp (del, sig) -> SigExp CSaudio)
> delayArrow = arrowUncurry delay1

Note that we have skipped using arr in many cases, mostly to ease
readability, since it was defined as id anyway.

Now let's define the signal we provided as an example in the previous
part (not the example given above, since it would complicate the
presentation without giving any extra insights on using the arrow
interface). This is our first try. Note that the instrument we 
define should actually be a function from a signal that is a list
of all the p-fields to an audio-rate signal. In a setting where
more powerful dependent types were available, we would have the
p-field list reflect its length, and allow only available p-fields
to be extracted out of it.

> signalArrow :: SigExp [CSnote] -> SigExp CSaudio
> signalArrow = let pch = Nth 4
>                   vel = Nth 5
>                   dur = Nth 3
>                   freq = pch >>> Prefix "cpspch"
>                   amp  = vel >>> Prefix "ampdb"
>                   envPar = ConstFloat 0.0 &&& dur &&& ConstFloat 1.0
>                   env  = envPar >>> (lineArrow CR)
>                   amp' = (amp &&& env) >>> (Infix "*")
>                   modAmp  = ConstFloat 5.0
>                   modFreq = ConstFloat 5.0 &&& dur >>> Infix "*"
>                   mod  = modAmp &&& modFreq >>> oscPureArrow
>                   freq' = freq &&& mod >>> Infix "+"
>                   inp  = amp' &&& freq'
>                   sig  = inp >>> oscPureArrow
>                   recfunc = res &&& res
>                             where res = second (delay1 (ConstFloat 0.311 Unit)) >>> Infix "+"
>                   recloop = loop Tar recfunc
>                   sig' = sig >>> recloop
>               in sig'

Not bad! Let's define some syntactic sugar to make this even more
readable.

> notePit :: SigExp [CSnote] -> SigExp CSnote
> notePit = Nth 4
> noteVel :: SigExp [CSnote] -> SigExp CSnote
> noteVel = Nth 4
> noteDur :: SigExp [CSnote] -> SigExp CSnote
> noteDur = Nth 3
> pchToHz :: forall a. CSType a => SigExp a -> SigExp a
> pchToHz  = Prefix "cpspch"
> ampToDb :: forall a. CSType a => SigExp a -> SigExp a
> ampToDb  = Prefix "ampdb"
> (><)    :: ( SigExp x -> SigExp y ) -> ( SigExp x -> SigExp y' ) -> ( SigExp x -> SigExp (y, y') )
> (><)     = (&&&)
> (<+>)   :: CSisSubtype a b => SigExp (a,b) -> SigExp b
> (<+>)    = Infix "+"
> (<*>)   :: CSisSubtype a b => SigExp (a,b) -> SigExp b
> (<*>)    = Infix "*"
> cnst    :: forall a. Float -> SigExp a -> SigExp CSconst
> cnst    = ConstFloat

> signalArrow' :: SigExp [CSnote] -> SigExp CSaudio
> signalArrow' = let freq = notePit >>> pchToHz
>                    amp  = noteVel >>> ampToDb
>                    env  = (cnst 0.0 >< noteDur >< cnst 1.0) >>> (lineArrow CR)
>                    amp' = amp >< env >>> (<*>)
>                    modAmp  = cnst 5.0
>                    modFreq = cnst 5.0 >< noteDur >>> (<*>)
>                    mod  =  modAmp >< modFreq >>> oscPureArrow
>                    freq' = freq >< mod >>> (<+>)
>                    sig  = amp' >< freq' >>> oscPureArrow
>                    recfunc = res >< res
>                              where res = second delayFunc >>> (<+>)
>                                    delayFunc = delay1 (cnst 0.311 Unit)
>                    recloop = loop Tar recfunc
>                    sig' = sig >>> recloop
>                in sig'


Now we would also like to use the actual arrow syntax to define
signals, but as we mentioned earlier, this is not possible (since we
use a redefined Arrow type class). It should not be too hard to change
the existing arrow preprocessor to handle our class though, provided
that we gave a type to each recursive signal we define (so that the
loop function can be called accordingly). Assuming that we have such
a thing, we could write code similar to the following:

Here we use a different definition for (><):
(><) = Pair

signalArrowSyn :: SigExp [CSnote] -> SigExp CSaudio
signalArrowSyn = proc inp -> do
                 rec  pch       <- Nth 4           -< inp
                      vel       <- Nth 5           -< inp
                      dur       <- Nth 3           -< inp
                      freq      <- pchToHz         -< pch
                      amp       <- ampToDb         -< vel
                      lineP1    <- cnst 0.0        -< inp
                      lineP3    <- cnst 1.0        -< inp
                      env       <- lineArrow CR    -< lineP1 >< dur >< lineP3
                      amp'      <- (<*>)           -< amp >< env
                      modAmp    <- cnst 5.0        -< inp
                      modFreqP1 <- cnst 5.0        -< inp
                      modFreq   <- (<+>)           -< modFreqP1 >< dur
                      mod       <- oscPureArrow    -< modAmp >< modFreq
                      freq'     <- (<+>)           -< freq >< mod
                      sig       <- oscPureArrow    -< amp' >< freq'
                      delayP1   <- cnst 0.311      -< inp
               [Tar]  sig'Del   <- delayArrow      -< delayP1 >< sig'
               [Tar]  sig'      <- (<+>)           -< sig >< sig'Del
                 returnA -< sig'


OK, now we have seen everything but how to actually compile our
enhanced SigExp's into CSound. That turns out to be an anything but
trivial task. The problem is that we can't directly compile into
SigExp's of the original HasSound, because the compilation wouldn't be
able to handle pairs. We need to support pairs at all levels, even as
variables in recursive signals, and that's hard to get under the
current HasSound. Thus, instead of compiling into the original
SigExp's, I took another approach: compile directly into CSound
orchestra files!  The code that does the actual translation from
SigExp's into CSound orchestra files follows. Note that this is not
general at all: it only creates the code for one instrument (it
doesn't even generate the header, it must be copy-and-paste'd into the
resulting code from a working CSound orchestra file with the same
tables), it doesn't handle comparisons or delay lines, etc. Also, it
is really hacky code and unreadable. Nevertheless, it works fine, at
least for the given examples.

First, we provide the testing functions.

> test = putStrLn . output

> pfieldInput :: Int -> SigExp [CSnote]
> pfieldInput n = foldl (\cur n-> Cons (PField n) cur) Nil [1..n]
> testArrow sf = putStrLn . output $ sf (pfieldInput 5)

You can use "test signal" and "testArrow signalArrow" or "testArrow
signalArrow'" to test the above examples.


Now the actual conversion code comes. We will not describe everything,
just the general idea of each function.

output: The main conversion function, it gets an audio-rate signal and
produces a string that is the CSound-score-representation of the
signal.

> output :: SigExp CSaudio -> [Char]
> output sig = "instr 1\n" ++ init ++ concat ops ++ "out " ++ var ++ "\nendin\n"
>              where (initL, ops, Leaf(_, var)) = convert "" $ canonicalize AR sig
>                    init = concat (map (++ (" init 0\n")) initL)

A binary tree type and some functions for it.

> data Tree a where
>     Leaf :: a -> Tree a
>     Node :: Tree a -> Tree a -> Tree a

> instance Functor Tree where
>     fmap f (Leaf a) = Leaf (f a)
>     fmap f (Node a b) = Node (fmap f a) (fmap f b)

> getLeaf  (Leaf a  ) = a
> getLeft  (Node a b) = a
> getRight (Node a b) = b
> treeFlatten (Leaf a) = [a]
> treeFlatten (Node a b) = treeFlatten a ++ treeFlatten b

getrate: Given an evalrate, find what the first character of a
variable of that rate must be.

> getrate :: forall a . EvalRate a -> String
> getrate AR = "a"
> getrate CR = "k"
> getrate NR = "i"

canonicalize: Given an audio signal of some rate, it encloses it in
the identity signal generator (which is "=" in CSound), if it's not
the result of a signal generator already. This is done so that the
resulting SigExp is forced to have an outer signal generator, and it
will thus be translated into a separate variable.

> canonicalize r (sig@(SigGen _ _ _)) = sig
> canonicalize r sig = SigGen "=" r [ Hide sig ]

canonicalizeAll: the same thing as canonicalize, only that it can
handle signals of pair type as well. It forces all primitive members
of an arbitrarily paired signal to be evaluated into separate
variables.

> canonicalizeAll :: forall a. AllTypes a -> SigExp a -> SigExp a
> canonicalizeAll (Tpair tA tB) (Pair a b) = Pair (canonicalizeAll tA a) (canonicalizeAll tB b)
> canonicalizeAll (Tunit) (Unit) = Unit
> canonicalizeAll r (sig@(SigGen _ _ _)) = sig
> canonicalizeAll Tar sig = SigGen "=" AR [ Hide sig ]
> canonicalizeAll Tcr sig = SigGen "=" CR [ Hide sig ]
> canonicalizeAll Tnr sig = SigGen "=" NR [ Hide sig ]

convert: the hacky conversion function. The first argument is a string that
describes the path taken to reach the current SigExp, and is used as an
easy way to generate unique variable names. The second argument is the SigExp
to be converted. It then produces three results: the first is a list of 
the names of the variables that must be initialized at the start of the
instrument (used for recursive variables). The second is a list of the
signal generator operations that the SigExp does. The third result is the simple
expression that can be used as the result of the signal anywhere else. For
example, a SigExp that is just a + b will have nothing as the second result,
and "a + b" as the third. A SigExp that is (SigGen "f" [a]) + b will have
["a1 f a"] as the second result, and "a1 + b" as the third.

The third result is actually more complex than the obvious thing, because
of the way we compile pairs. It will actually be a tree with two values
at each leaf. Each leaf represents a different primitive SigExp that is
part of the arbitrary pair we are converting. The values at each leaf are
one string that describes the rate of the signal, and one string that
is actually the simple expression that can be used for that signal.

> convert :: forall a . [Char] -> SigExp a -> ([[Char]], [[Char]], Tree ([Char], [Char]))
> convert path Unit             = error "converting unit?"
> convert path (ConstFloat x _) = ([], [], Leaf $ ("i", show x))
> convert path (ConstInt   x _) = ([], [], Leaf $ ("i", show x))
> convert path (TableNumber x _)= ([], [], Leaf $ ("i", show x))
> convert path (PField i)       = ([], [], Leaf $ ("i", "p" ++ show i))
> convert path (Infix f a)      = (initA ++ initB, listA ++ listB, Leaf $ (rate, "(" ++ convA ++ " " ++ f ++ " " ++ convB ++ ")"))
>                                  where (initA, listA, Leaf (rateA, convA)) = convert (path ++ "1_") (First a)
>                                        (initB, listB, Leaf (rateB, convB)) = convert (path ++ "2_") (Second a)
>                                        rate = calcrate rateA rateB
>                                        calcrate "a" _ = "a"
>                                        calcrate _ "a" = "a"
>                                        calcrate "k" _ = "k"
>                                        calcrate _ "k" = "k"
>                                        calcrate _ _   = "i"
> convert path (Prefix f a)     = (initA, listA, Leaf $ (rateA, f ++ "(" ++ convA ++ ")"))
>                                  where (initA, listA, Leaf (rateA, convA)) = convert (path ++ "1_") a
> convert path (SigGen f a l)   = (initL, paramList ++ [thisOp ++ "\n"], Leaf (getrate a, newVar))
>                                  where paramConv = [ convert (path ++ show n ++ "_") p | (n,Hide p) <- zip [1..] l ]
>                                        paramList = concat . (map (\(x,y,z)->y)) $ paramConv
>                                        params    = map (\(x,y,z)->snd (getLeaf z)) $ paramConv
>                                        initL     = concat . (map (\(x,y,z)->x)) $ paramConv
>                                        evalChar  = getrate a
>                                        newVar    = evalChar ++ path ++ "0"
>                                        thisOp    = newVar ++ " " ++ f ++ " " ++ convertList params
>                                        convertList []         = ""
>                                        convertList (hd : []) = hd
>                                        convertList (hd : tl) = hd ++ ", " ++ convertList tl
> convert path (Recurse varT f)   = (resInit ++ updInit ++ newVarsFlat, resParam ++ updParam, resRes)
>                                    where newVarsNes = getVarNames (path ++ "rec") varT
>                                          newVarsFlat = treeFlatten newVarsNes
>                                          (result, updated) = f (Var newVarsNes)
>                                          (updInit, updParam, updRes) = convert (path ++ "rec") (canonicalizeAll varT updated)
>                                          (resInit, resParam, resRes) = convert (path ++ "res") result
>                                          getVarNames :: forall a . String -> AllTypes a -> Tree String
>                                          getVarNames path Tar = Leaf ("a" ++ path ++ "0")
>                                          getVarNames path Tcr = Leaf ("k" ++ path ++ "0")
>                                          getVarNames path Tnr = Leaf ("i" ++ path ++ "0")
>                                          getVarNames path Tunit = Leaf ""
>                                          getVarNames path (Tpair a b) = Node (getVarNames (path ++ "p1_") a) (getVarNames (path ++ "p2_") b)

> convert path (Var x)         = ([], [], fmap (\x -> ([head x], x)) x)

> convert path (Pair x y)      = (initX ++ initY, opX ++ opY, Node xRes yRes)
>                                 where (initX, opX, xRes) = convert (path ++ "p1_") x
>                                       (initY, opY, yRes) = convert (path ++ "p2_") y
> convert path (First (Pair x y))  = convert (path ++ "p1_") x
> convert path (Second (Pair x y)) = convert (path ++ "p2_") y
> convert path (First (First (Pair (Pair x y) z))) = convert (path ++ "p1_p1_") x
> convert path (Second (First (Pair (Pair x y) z))) = convert (path ++ "p1_p2_") y
> convert path (First (Second (Pair x (Pair y z)))) = convert (path ++ "p2_p1_") y
> convert path (Second (Second (Pair x (Pair y z)))) = convert (path ++ "p2_p2_") z
> convert path (First a)       = (init, ops, getLeft res)
>                                 where (init, ops, res) = convert path a
> convert path (Second a)      = (init, ops, getRight res)
>                                 where (init, ops, res) = convert path a
> convert path (Nth 1 (Cons x _))   = convert (path ++ "l1_") x
> convert path (Nth n (Cons x y))   = convert (path ++ "l2_") (Nth (n-1) y)
> convert path (Nth n _)            = error ("reading past the end of a list" ++ path)

