Fundamentals of Computer Music Final Project
Antonis Stampoulis (antonios.stampoulis@yale.edu)

In the first part of my project, I did an investigation of the
feasibility of enforcing stricter type checking to HasSound
expressions using GADTs. It should be noted that the following code is
meant as proof-of-concept, and is not ready to become part of
HasSound.

The general idea is that we want the types of SigExps to reflect
the rate they're computed at by CSound, so that operations on signals
are only permitted to be applied to arguments of the correct rate.

This module works as following: it provides its own SigExp datatype,
which gets translated after the typechecking phase into the normal
Orchestra.SigExp type. The user of this module should use only the
provided constructors to produce the desired signal expression, and
not the constructors in Orchestra.SigExp. Then, conversion to
the original SigExp type must be done using the convert function.

Since the Orchestra.lhs file doesn't export everything we need here,
we need to modify it for this file to compile: we must add the
functions prefixTerm, infixTerm, operator, comparisonTerm and pField
to the export list of Orchestra.lhs.


First we define the modules that we'll need.


> module OrchestraStrict where

> import qualified Haskore.Interface.CSound.Orchestra    as Orchestra
> import Haskore.Interface.CSound.Score        as Score

> import Haskore.Interface.CSound.Render (playCS)

> import qualified Haskore.Performance         as Performance
> import qualified Haskore.Performance.Context as Context
> import qualified Haskore.Performance.Player  as Player

> import Haskore.Music                         as Music
> import Haskore.Melody                        as Melody
> import qualified Haskore.Melody.Standard     as StdMelody
> import qualified Haskore.Music.Rhythmic      as RhyMusic



First let's define the types for CSound. Ideally, we'd have an
extensible-kind type-system, where we could define a separate kind for
CSound. Since there's no such thing in Haskell yet, we just define
normal void datatypes.

> data CSconst
> data CSnote
> data CScontrol
> data CSaudio

Then we implement the subtyping rules for CSound using the type class
CSisSubtype. This type class takes two arguments, and has an instance
for every pair whose first element is a CSound-subtype of the other.
The subtyping rules are really straightforward: a rate is a subtype
of itself and all the higher rates.

> class CSisSubtype a b
> instance CSisSubtype CSconst CSconst
> instance CSisSubtype CSconst CSnote
> instance CSisSubtype CSconst CScontrol
> instance CSisSubtype CSconst CSaudio
> instance CSisSubtype CSnote  CSnote
> instance CSisSubtype CSnote  CScontrol
> instance CSisSubtype CSnote  CSaudio
> instance CSisSubtype CScontrol CScontrol
> instance CSisSubtype CScontrol CSaudio
> instance CSisSubtype CSaudio   CSaudio

Now we'd like to define a type class CScalcSubtype with three arguments,
to calculate the largest type of two types. That is, it should have an
instance for every triplet where the third element is the largest type
of the first two elements. We would define it as follows:

class CScalcSubtype a b c
instance (CSisSubtype a b) => CScalcSubtype a b b
instance (CSisSubtype b a) => CScalcSubtype a b a

Only, this isn't very helpful, because the Haskell constraint solver
isn't powerful enough to handle the cases we need -- that's probably
because of bad interaction with GADTs. Ideally, we'd want to have a
way to define functions at the type level, so an implementation of
CScalcSubtype would be straightforward and would always work. Since
this isn't available, we try to simulate it with type classes, but
that doesn't restrict us to total functions, hence the problems.

Let's now proceed to the main section: definition of SigExp's and
related types. First, a really easy one: CSound functions are
represented as strings.

> type Function = String

The second is a strict EvalRate, that reflects its value in its
type. This is the first GADT we define, and it is evident that
it is a form of (fully) dependent type.

> data EvalRate a where
>     AR :: EvalRate CSaudio
>     CR :: EvalRate CScontrol
>     NR :: EvalRate CSnote

We'll now define the SigExp datatype. It's partially dependent: only
on the rate of the resulting signal. Note that this doesn't include
everything that is defined in the original HasSound, but this was done
to ease implementation -- the other parts could be added too. Here we
have included the most essential part of SigExp's : normal
expressions, signal generators, recursive signals and conditionals.

We will describe each constructor we define to help understanding
this.

> data SigExp a where

Constants and table numbers are evaluated in fixed rate. They can
be implicitly cast to any other rate.

>     ConstFloat   :: Float -> SigExp CSconst
>     ConstInt     :: Int   -> SigExp CSconst
>     TableNumber  :: Int   -> SigExp CSconst

PFields are evaluated at note rate.

>     PField       :: Int   -> SigExp CSnote

Applying a prefix operator to a signal yields a signal of the
same type.

>     Prefix       :: Function -> SigExp a -> SigExp a

Applying an infix operator to two signals yields a signal of
the type that is the largest of the two input signals' types.
The CScalcSubtype function we mentioned before would be put at
use here, but it wouldn't be of help: Haskell's constraint solver
cannot derive the CSound-type of the resulting signal, so the user
has to specify it. Instead, we chose another solution: to have the
Infix constructor require that the first operand's type is of smallest
or equal type to the second operand's. This is easily implemented using
CSisSubtype: the first signal's type must be a subtype of the second,
therefore the resulting signal has the same type as the second signal.

>     Infix        :: CSisSubtype a b => Function -> SigExp a -> SigExp b -> SigExp b

Similar handling is done for the conditionals: they accept a boolean term
at a certain rate, and two signals of equal or higher rate. The two signals
must be at the same rate, and so will be the resulting signal.

>     Conditional  :: CSisSubtype a b => BooleanTerm a -> SigExp b -> SigExp b -> SigExp b

The cast constructor is provided in case an explicit cast is needed from one
type to the other. The rate conversion can only be done towards higher rates!

>     Cast         :: CSisSubtype a b => SigExp a -> SigExp b

The SigGen constructor is probably the most interesting. It gets a function
and the evaluation rate of the result of the function. The resulting signal
will be of the same rate. But what's the type of the list of parameters? The
list members do not have all the same type! This is because different rates
are now reflected in the types of signals, so signals of different rates are
not of the same type. We need a type that permits a SigExp of any rate to
be part of the list. To achieve this, we need a type that hides the parameter
to the SigExp type constructor, but requires that such a type exists. Evidently,
we must use an existential type for this. Haskell doesn't yet support special
syntax for existential types -- if that was the case, we would simple write
[ exists t. SigExp t ] for the type of the parameter list. Instead, we
define an extra datatype, SigExpAny, that is actually an existential type. It
provides the sole constructor Hide, which produces a SigExp with the type
parameter hidden. We therefore use [ SigExpAny ] as the type of the parameter
list. Note that no extra type checking on the arguments is being performed!
Our wrapper functions will have to have stricter types than this (that reflect
the rate restrictions), and use the Hide constructor to make parameters part
of the parameter list.

>     SigGen       :: Function -> EvalRate a -> [ SigExpAny ] -> SigExp a

The Recurse constructor takes as argument the evaluation rate of the recursive
signal, and the function that given the previous value of the recursive signal,
provides the current. The result is of the same rate as the recursive signal.

>     Recurse      :: EvalRate a -> (SigExp a -> SigExp a) -> SigExp a

The Unsafe constructor is to be used internally only. It gets an evaluation
rate and an original-HasSound SigExp, and produces a SigExp of the specified
rate. It is unsafe because we must trust the user for the actual rate of the
original SigExp.

>     Unsafe       :: EvalRate a -> Orchestra.SigExp -> SigExp a

SigExp constructors are over!

This is the boolean term constructor. It's a GADT too.

> data BooleanTerm a where
>     Operator     :: CSisSubtype a b => Function -> BooleanTerm a -> BooleanTerm b -> BooleanTerm b
>     Comparison   :: CSisSubtype a b => Function -> SigExp a -> SigExp b -> BooleanTerm b

This is the existential type SigExpAny we mentioned above.

> data SigExpAny where
>     Hide :: forall a. SigExp a -> SigExpAny

Some syntactic sugar: define constraints CSaudio a similar
to CSisSubtype a CSaudio, etc.

> class CSConst   a
> instance CSConst CSconst

> class CSNote    a
> instance CSNote CSconst
> instance CSNote CSnote

> class CSControl a
> instance CSControl CSconst
> instance CSControl CSnote
> instance CSControl CScontrol

> class CSAudio   a
> instance CSAudio CSconst
> instance CSAudio CSnote
> instance CSAudio CScontrol
> instance CSAudio CSaudio

This is the function that converts a SigExp defined here to a Mono
output from Orchestra. Internally, it converts to an Orchestra.SigExp,
and passes that into Orchestra.Mono.

> toHasSound :: forall a . SigExp a -> Orchestra.Mono
> toHasSound = Orchestra.Mono . convert
>     where convert :: forall a. SigExp a -> Orchestra.SigExp
>           convert (ConstFloat x)  = Orchestra.constFloat  x
>           convert (ConstInt x)    = Orchestra.constInt    x
>           convert (TableNumber x) = Orchestra.tableNumber x
>           convert (PField x)      = Orchestra.pField      x
>           convert (Infix  x a b)  = Orchestra.infixTerm   x (convert a) (convert b)
>           convert (Prefix x a)    = Orchestra.prefixTerm  x (convert a)
>           convert (SigGen f r l)  = Orchestra.sigGen f (convertRate r) 1 (map (convertAny) l)
>           convert (Unsafe r w)    = w

The most interesting thing is the translation of recursive signals. We
use the rec function provided by Orchestra to implement this. The
function we pass to it is a function that takes an Orchestra.SigExp, and produces
an Orchestra.SigExp. It uses the rate specifier and the unsafe constructor to
create a strict SigExp which it can use as an argument to the provided function
(since it accepts and produces strict SigExp's). The strict SigExp produced is
then translated into an Orchestra.SigExp.

>           convert (Recurse r f)   = Orchestra.rec (\x -> convert (f (Unsafe r x)))
>           convert (Conditional b t f) = Orchestra.ifthen (convertBoolean b) (convert t) (convert f)
>           convert (Cast a)        = convert a

This is an interesting function: convertAny takes a signal of any rate
(using the existential type) and converts it to an
Orchestra.SigExp. Matching against the Hide constructor is like
performing the unpack operation of existentials: the hidden type can
get used in the expression scoped inside it (as is done here, because
convert expects a strict SigExp that includes the type), but cannot
escape the unpack operation!

>           convertAny :: SigExpAny -> Orchestra.SigExp
>           convertAny (Hide a) = convert a
>           
>           convertRate :: forall a. EvalRate a -> Orchestra.EvalRate
>           convertRate AR = Orchestra.AR
>           convertRate CR = Orchestra.CR
>           convertRate NR = Orchestra.NR
>
>           convertBoolean :: forall a. BooleanTerm a -> Orchestra.Boolean
>           convertBoolean (Operator f a b) = Orchestra.operator f (convertBoolean a) (convertBoolean b)
>           convertBoolean (Comparison f a b) = Orchestra.comparisonTerm f (convert a) (convert b)


Now let's try it out!

> pureToneTable = TableNumber pureToneTN

We define the strict oscillator function: it gets an evaluation rate, a
table (which should be of constant rate), and an amplitude and frequency (of
up to audio rate), and produces a signal of the specified rate. Note the
usage of the Hide constructor as the equivalent of the pack operation of
existentials. We do not care that no further type checking is done on the
sigGen's parameters (after we have hidden the rates from the type), because
osc's type is strict enough so that it does all the necessary tests.

> osc :: (CSAudio amp, CSAudio freq, CSConst table) =>
>        EvalRate a -> SigExp table -> SigExp amp -> SigExp freq -> SigExp a
> osc rate table amp freq = SigGen "oscil" rate [ Hide amp, Hide freq, Hide table ]

Similarly we define oscPure that operates on the pure-tone table, with the
appropriate rate restrictions.

> oscPure :: (CSAudio amp, CSAudio freq) =>
>            SigExp amp -> SigExp freq -> SigExp CSaudio
> oscPure = osc AR pureToneTable

The line sigGen, used to generate ramps etc. Note that all its parameters
must be up to note rate! (We use line1 as the name to avoid a conflict)

> line1 :: (CSNote ia, CSNote idur1, CSNote ib) =>
>          EvalRate a -> SigExp ia -> SigExp idur1 -> SigExp ib -> SigExp a
> line1 rate start dur end = SigGen "line" rate [ Hide start, Hide dur, Hide end ]

Similarly we define the delay sigGen. We use delay1 to avoid a conflict in
this case too.

> delay1 :: (CSAudio sig, CSNote del) =>
>           SigExp del -> SigExp sig -> SigExp CSaudio
> delay1 del sig = SigGen "delay" AR [ Hide sig, Hide del ]

Now we are ready to define a strict signal, that has its rate
specified and checked by Haskell's typechecker! Note the use of
recursion.

> signal :: SigExp CSaudio
> signal =
>     let freq = Prefix "cpspch" (PField 4)
>         amp  = Prefix "ampdb"  (PField 5)
>         env  = line1 CR (ConstFloat 0.0) (PField 3) (ConstFloat 1.0)
>         amp' = Infix "*" amp env
>         freq' = Infix "+" freq (oscPure (ConstFloat 5.0) (Infix "*" (ConstFloat 5.0) (PField 3)))
>         sig  = oscPure amp' freq'
>         sig' = Recurse AR (\x -> Infix "+" sig (delay1 (ConstFloat 0.311) x))
>     in  sig'

Now let's try it out! Just use 'test tut' at the ghci prompt.

> tut = mkTut score1 [toHasSound signal]

We define a conditional signal too, to test typechecking in the
case of boolean terms, etc. The output it produces doesn't have
any meaning! Note the use of the Cast constructor.

> condsignal :: SigExp CSaudio
> condsignal =
>     let nr = PField 4
>         cr = line1 CR (ConstFloat 0.0) (PField 3) (ConstFloat 1.0)
>         ar = oscPure nr nr
>     in
>     Conditional (Operator "&&" (Comparison "<" nr nr) (Comparison "<" (Cast nr :: SigExp CScontrol) cr)) ar (Cast nr)

Example of a signal that shouldn't work:

signal_error :: SigExp CSaudio
signal_error =
  let freq = Prefix "cpspch" (PField 4)
      amp  = Prefix "ampdb"  (PField 5)
      tst  = oscPure amp freq
      env  = line1 CR tst (PField 3) (ConstFloat 1.0)
      amp' = Infix "*" amp env
      sig  = oscPure amp' freq
  in  sig

Here the first argument of line is not at note rate, so this should
report an error! Turning the above into a code fragment and compiling
this file again, the Haskell compiler will issue a type error, because
the rate of the tst signal is not a subtype of the note rate.

This of course should be extended to include the other things
supported by the original HasSound. But, there is a serious problem:
providing instances for Num, Fractional and Floating, as is done in
the original HasSound, is not possible! Actually, we have to provide
separate instances for all the different rates we support; that is, a
separate instance for SigExp CSconst, SigExp CSnote, SigExp CScontrol
and SigExp CSaudio. This wouldn't be too much of a problem if it would
work, but it won't! The reason is that using the resulting operators,
we couldn't for example add two signals of different rate
together. Defining these operators for the existential type would
compromise all the work we have done so far, since no extra type
checking would be done. The user would thus have to provide explicit
cast operators when combining two signals of different rate together.

-------------------------------------------------------------------
The following declarations provide a test tune and a test function
that we can use to try out our signal expressions.  They are copied
verbatim from the prologue my homework for exercise 2. No changes
where needed in this part, so you can skip it!
-------------------------------------------------------------------

> pureToneTN :: Score.Table
> pureToneTN = 1
> pureTone :: Score.Statement
> pureTone = Score.simpleSine pureToneTN
> 
> type InstrNum = Int
> 
> type Instrument = ([Float], InstrNum)
> 
> instr1, instr2 :: InstrNum
> instr1 = 1
> instr2 = 2
> 
> hdr :: Orchestra.Header
> hdr = (44100, 4410)
> 
> type TutMelody = Melody.T TutAttr
> 
> data TutAttr = TutAttr {attrVelocity   :: Rational,
>                         attrParameters :: [Float]}
> type Drum = ()
> 
> musicFromMelody :: InstrNum -> TutMelody -> RhyMusic.T Drum Instrument
> musicFromMelody instrId = Music.mapNote (
>             \(Melody.Note (TutAttr vel params) p) ->
>              RhyMusic.Note vel (RhyMusic.Tone (params, instrId) p))
> 
> scoreFromMelody :: InstrNum -> TutMelody -> Score.T
> scoreFromMelody i m = 
>    (Score.fromRhyPerformanceMap
>        (error "no drum map defined") id) $
>    (Performance.fromMusic Player.fancyMap 
>        (Context.setDur 1 Context.deflt)) $ musicFromMelody i m
> 
> type Example out = (Orchestra.T out, Score.T)
> 
> mkTut :: Orchestra.Output out => Score.T -> [out] -> Example out
> mkTut score oes = 
>    let o = Orchestra.Cons hdr (map (\(i, oe) -> Orchestra.InstrBlock i 0 oe []) (zip [1..] oes))
>     in (o, score)
> 
> test :: Orchestra.Output out => Example out -> IO ()
> test = uncurry playCS
> 
> tune1 :: TutMelody
> tune1 = Music.line (map ($ TutAttr 1.4 [])
>                     [ c 2 wn, b 1 wn, a 1 wn,
>                       g 1 wn, f 1 wn, e 1 wn,
>                       f 1 wn, g 1 wn ])
> 
> score1 = pureTone : scoreFromMelody instr1 tune1
